print("import: 'keras'")
import keras

print("import: 'keras.backend'")
import keras.backend

print("import: 'keras.datasets'")
import keras.datasets

print("import: 'keras.engine'")
import keras.engine

print("import: 'keras.layers'")
import keras.layers

print("import: 'keras.legacy'")
import keras.legacy

print("import: 'keras.preprocessing'")
import keras.preprocessing

print("import: 'keras.utils'")
import keras.utils

print("import: 'keras.wrappers'")
import keras.wrappers

